@@ -197,12 +197,52 @@ message.
         with self.assertRaises(DiscardMessage) as cm:
             mime_delete.dispose(self._mlist, self._msg, {}, 'preserved')
         self.assertEqual(cm.exception.message, 'preserved')
-        # There should be no messages in the 'bad' queue.
+        # There should be a message in the 'bad' queue.
         items = get_queue_messages('bad', expected_count=1)
         message = items[0].msg
         self.assertEqual(message['subject'], 'A disposable message')
         self.assertEqual(message['message-id'], '<ant>')
 
+    @configuration('mailman', filtered_messages_are_preservable='yes')
+    def test_preserved_message_has_content(self):
+        msg = mfs("""\
+From: anne@example.com
+To: test@example.com
+Subject: Testing preserved message has content
+Message-ID: <ant>
+MIME-Version: 1.0
+Content-Type: multipart/mixed; boundary="AAAA"
+
+--AAAA
+Content-Type: text/bogus; charset="utf-8"
+Content-Transfer-Encoding: quoted-printable
+
+Let=E2=80=99s also consider
+
+--AAAA
+Content-Type: text/other; charset="utf-8"
+Content-Transfer-Encoding: quoted-printable
+
+Let=E2=80=99s also consider
+
+--AAAA--
+""")
+        self._mlist.filter_content = True
+        self._mlist.filter_action = FilterAction.preserve
+        self._mlist.pass_types = ['multipart', 'text/plain']
+        with self.assertRaises(DiscardMessage) as cm:
+            mime_delete.process(self._mlist, msg, {})
+        self.assertEqual(cm.exception.message,
+                         'After content filtering, the message was empty')
+        # There should be a message in the 'bad' queue.
+        items = get_queue_messages('bad', expected_count=1)
+        message = items[0].msg
+        self.assertEqual(message['subject'],
+                         'Testing preserved message has content')
+        self.assertEqual(message['message-id'], '<ant>')
+        # The message should be multipart with two subparts.
+        self.assertEqual(len(message.get_payload()), 2)
+
     def test_bad_action(self):
         # This should never happen, but what if it does?
         # FilterAction.accept, FilterAction.hold, and FilterAction.defer are
@@ -619,5 +659,140 @@ Content-Type: image/jpeg
 Content-Type: video/mp4
     Name: My_video
 
+Replaced multipart/alternative part with first alternative.
+""")
+
+    def test_report_mixed(self):
+        # Hit all the pass and filter conditions for reporting
+        self._mlist.pass_extensions = ['txt']
+        self._mlist.pass_types = ['text', 'application', 'multipart', 'image']
+        msg = mfs("""\
+From: anne@example.com
+To: test@example.com
+Subject: Testing mixed extension
+Message-ID: <ant>
+MIME-Version: 1.0
+Content-Type: multipart/mixed; boundary="AAAA"
+
+--AAAA
+Content-Type: multipart/alternative; boundary="BBBB"
+
+--BBBB
+Content-Type: text/plain; charset="utf-8"
+
+Plain text
+
+--BBBB
+Content-Type: text/html; charset="utf-8"
+
+HTML text
+
+--BBBB--
+--AAAA
+Content-Type: application/octet-stream; name="test.xlsX"
+Content-Disposition: attachment; filename="test.xlsX"
+
+spreadsheet
+
+--AAAA
+Content-Type: application/octet-stream; name="test.exe"
+Content-Disposition: attachment; filename="test.exe"
+
+executable
+
+--AAAA
+Content-Type: image/jpeg; name="My_image"
+Content-Disposition: inline; filename="My_image"
+
+image
+
+--AAAA
+Content-Type: video/mp4; name="My_video"
+Content-Disposition: inline; filename="My_video"
+
+video
+
+--AAAA--
+""")
+        process = config.handlers['mime-delete'].process
+        with dummy_script('report'):
+            process(self._mlist, msg, {})
+        self.assertIn('multipart/mixed', msg['content-type'])
+        self.assertEqual(len(msg.get_payload()), 3)
+        self.assertIn('text/plain', msg.get_payload(0).get_content_type())
+        self.assertIn('image/jpeg', msg.get_payload(1).get_content_type())
+        self.assertEqual(msg.get_payload(2).get_payload(decode=True), b"""\
+
+___________________________________________
+Mailman's content filtering has removed the
+following MIME parts from this message.
+
+Content-Type: application/octet-stream
+    Name: test.xlsX
+
+Content-Type: application/octet-stream
+    Name: test.exe
+
+Content-Type: video/mp4
+    Name: My_video
+
+Replaced multipart/alternative part with first alternative.
+""")
+
+    def test_report_other_multi(self):
+        # Hit all the pass and filter conditions for reporting
+        self._mlist.pass_extensions = ['txt']
+        self._mlist.pass_types = ['text', 'application', 'multipart']
+        msg = mfs("""\
+From: anne@example.com
+To: test@example.com
+Subject: Testing mixed extension
+Message-ID: <ant>
+MIME-Version: 1.0
+Content-Type: multipart/signed; boundary="AAAA"
+
+--AAAA
+Content-Type: multipart/alternative; boundary="BBBB"
+
+--BBBB
+Content-Type: text/plain; charset="utf-8"
+
+Plain text
+
+--BBBB
+Content-Type: application/pkcs7-signature
+
+HTML text
+
+--BBBB--
+--AAAA
+Content-Type: application/pkcs7-signature
+
+Signature
+--AAAA
+Content-Type: application/octet-stream; name="test.xlsX"
+Content-Disposition: attachment; filename="test.xlsX"
+
+spreadsheet
+
+--AAAA--
+""")
+        process = config.handlers['mime-delete'].process
+        with dummy_script('report'):
+            process(self._mlist, msg, {})
+        self.assertIn('multipart/mixed', msg['content-type'])
+        self.assertEqual(len(msg.get_payload()), 2)
+        self.assertIn('multipart/signed',
+                      msg.get_payload(0).get_content_type())
+        self.assertIn('text/plain', msg.get_payload(1).get_content_type())
+        self.assertEqual(msg.get_payload(1).get_payload(decode=True), b"""\
+
+___________________________________________
+Mailman's content filtering has removed the
+following MIME parts from this message.
+
+Content-Type: application/octet-stream
+    Name: test.xlsX
+
 Replaced multipart/alternative part with first alternative.
 """)
