@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2022 by the Free Software Foundation, Inc.
+# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
 #
 # This file is part of GNU Mailman.
 #
@@ -22,7 +22,7 @@ import datetime
 
 from email.utils import make_msgid
 from lazr.config import as_boolean, as_timedelta
-from mailman.app.bounces import _ProbePendable, PENDABLE_LIFETIME, send_probe
+from mailman.app.bounces import _ProbePendable, send_probe
 from mailman.app.membership import delete_member
 from mailman.app.notifications import (
     send_admin_disable_notice,
@@ -104,7 +104,7 @@ class BounceProcessor:
         pendable = _ProbePendable(
             _mod_message_id=msg.get('message-id'))
         getUtility(IPendings).add(
-            pendable, lifetime=as_timedelta(PENDABLE_LIFETIME))
+            pendable, lifetime=as_timedelta(config.mailman.dsn_lifetime))
         event = BounceEvent(mlist.list_id, email, msg, where)
         store.add(event)
         return event
@@ -234,6 +234,9 @@ class BounceProcessor:
                 send_probe(member, msg=msg, message_id=event.message_id)
                 action = 'sending probe'
             else:
+                # Bounce score is reset by sending a probe but needs to be
+                # reset here too.  See #1061.
+                member.bounce_score = 0
                 self._disable_delivery(mlist, member, event)
                 action = 'disabling delivery'
             log.info(
